@@ -61,6 +61,10 @@ match_theme_image (GtkStyle       *style,
 	  match_data->state != image->match_data.state)
 	continue;
 
+      if ((flags & THEME_MATCH_POSITION) &&
+	  match_data->position != image->match_data.position)
+	continue;
+
       if ((flags & THEME_MATCH_SHADOW) &&
 	  match_data->shadow != image->match_data.shadow)
 	continue;
@@ -77,10 +81,22 @@ match_theme_image (GtkStyle       *style,
 	  match_data->gap_side != image->match_data.gap_side)
 	continue;
 
-      if (image->match_data.detail &&
-	  (!match_data->detail ||
-	   strcmp (match_data->detail, image->match_data.detail) != 0))
-      continue;
+      /* simple pattern matching for (treeview) details
+       * in gtkrc 'detail = "*_start"' will match all calls with detail ending
+       * with '_start' such as 'cell_even_start', 'cell_odd_start', etc.
+       */
+      if (image->match_data.detail)
+        {
+          if (!match_data->detail)
+            continue;
+          else if (image->match_data.detail[0] == '*')
+            {
+              if (!g_str_has_suffix (match_data->detail, image->match_data.detail + 1))
+                continue;
+            }
+          else if (strcmp (match_data->detail, image->match_data.detail) != 0)
+            continue;
+        }
 
       return image;
     }
@@ -89,10 +105,36 @@ match_theme_image (GtkStyle       *style,
 }
 
 static GdkBitmap *
-get_window_for_shape (ThemeImage *image, GdkWindow *window, GtkWidget *widget)
+get_window_for_shape (ThemeImage *image, GdkWindow *window, GtkWidget *widget,
+		      gint x, gint y, gint width, gint height)
 {
-  if (image->background_shaped)
+  /* It's not a good idea to set a shape mask when painting on anything but
+   * widget->window, not only does the shape mask get wrongly offset but also
+   * causes re-exposing the widget, and we'll re-apply the shape mask, ad
+   * infinitum.
+   *
+   * Noticed when GtkMenu was changed to do two paints, the other one being on
+   * ->bin_window (http://bugzilla.gnome.org/show_bug.cgi?id=169532)
+   */
+  if (image->background_shaped &&
+      widget && window == widget->window)
     {
+      gint window_width;
+      gint window_height;
+      /* It also is not a good idea to apply shape mask when filling a window
+       * partially.
+       *
+       * For example GtkMenu paints/painted the arrow backgrounds with exactly
+       * the same parameters as full window background - except for the
+       * geometry. (http://bugzilla.gnome.org/show_bug.cgi?id=404571)
+       */
+      if (x != 0 || y != 0)
+        return NULL;
+
+      gdk_drawable_get_size (window, &window_width, &window_height);
+      if (width != window_width || height != window_height)
+        return NULL;
+
       if (GTK_IS_MENU (widget))
 	return gtk_widget_get_parent_window (widget);
       else if (GTK_IS_WINDOW (widget))
@@ -102,6 +144,77 @@ get_window_for_shape (ThemeImage *image, GdkWindow *window, GtkWidget *widget)
   return NULL;
 }
 
+static void
+check_child_position (GtkWidget      *child,
+		      GList          *children,
+		      ThemeMatchData *match_data)
+{
+  GtkWidget *left;
+  GtkWidget *right;
+  GtkWidget *top;
+  GtkWidget *bottom;
+  GList *l;
+
+  left = right = top = bottom = child;
+
+  for (l = children; l != NULL; l = l->next)
+    {
+      GtkWidget *widget = l->data;
+
+      if (!GTK_WIDGET_DRAWABLE (widget))
+	continue;
+      
+      /* XXX Should we consider the lower right corner instead, for
+       * right/bottom? */
+
+      if (left->allocation.x > widget->allocation.x)
+	left = widget;
+      if (right->allocation.x < widget->allocation.x)
+	right = widget;
+      if (top->allocation.y > widget->allocation.y)
+	top = widget;
+      if (bottom->allocation.y < widget->allocation.y)
+	bottom = widget;
+    }
+
+  match_data->flags |= THEME_MATCH_POSITION;
+  match_data->position = 0;
+  if (left == child)
+    match_data->position |= THEME_POS_LEFT;
+  if (right == child)
+    match_data->position |= THEME_POS_RIGHT;
+  if (top == child)
+    match_data->position |= THEME_POS_TOP;
+  if (bottom == child)
+    match_data->position |= THEME_POS_BOTTOM;
+}
+
+static void
+check_buttonbox_child_position (GtkWidget *child, ThemeMatchData *match_data)
+{
+  GList *children = NULL;
+  GList *l;
+  GtkWidget *bbox;
+  gboolean secondary;
+
+  g_assert (GTK_IS_BUTTON_BOX (child->parent));
+  bbox = child->parent;
+
+  secondary = gtk_button_box_get_child_secondary (GTK_BUTTON_BOX (bbox), child);
+
+  for (l = GTK_BOX (bbox)->children; l != NULL; l = l->next)
+    {
+      GtkBoxChild *child_info = l->data;
+      GtkWidget *widget = child_info->widget;
+
+      if (child_info->is_secondary == secondary && GTK_WIDGET_DRAWABLE (widget))
+	children = g_list_prepend (children, widget);
+    }
+
+  check_child_position (child, children, match_data);
+  g_list_free (children);
+}
+
 static gboolean
 draw_simple_image(GtkStyle       *style,
 		  GdkWindow      *window,
@@ -109,25 +222,20 @@ draw_simple_image(GtkStyle       *style,
 		  GtkWidget      *widget,
 		  ThemeMatchData *match_data,
 		  gboolean        draw_center,
-		  gboolean        allow_setbg,
 		  gint            x,
 		  gint            y,
 		  gint            width,
 		  gint            height)
 {
+  gboolean maemo_position_theming;
   ThemeImage *image;
-  gboolean setbg = FALSE;
   
   if ((width == -1) && (height == -1))
-    {
-      gdk_window_get_size(window, &width, &height);
-      if (allow_setbg)
-      	setbg = TRUE;
-    }
+    gdk_drawable_get_size (window, &width, &height);
   else if (width == -1)
-    gdk_window_get_size(window, &width, NULL);
+    gdk_drawable_get_size (window, &width, NULL);
   else if (height == -1)
-    gdk_window_get_size(window, NULL, &height);
+    gdk_drawable_get_size (window, NULL, &height);
 
   if (!(match_data->flags & THEME_MATCH_ORIENTATION))
     {
@@ -138,6 +246,30 @@ draw_simple_image(GtkStyle       *style,
       else
 	match_data->orientation = GTK_ORIENTATION_HORIZONTAL;
     }
+
+  /* Check for maemo-position-theming to update the position data */
+  if (widget && widget->parent &&
+      gtk_widget_class_find_style_property (GTK_WIDGET_GET_CLASS (widget->parent),
+                                            "maemo-position-theming"))
+    {
+      gtk_widget_style_get (widget->parent,
+			    "maemo-position-theming", &maemo_position_theming,
+			    NULL);
+      if (maemo_position_theming)
+	{
+	  if (GTK_IS_BUTTON_BOX (widget->parent))
+	    check_buttonbox_child_position (widget, match_data);
+	  else
+	    {
+              /* Generic code for other kinds of containers */
+	      GList *children;
+
+	      children = gtk_container_get_children (GTK_CONTAINER (widget->parent));
+	      check_child_position (widget, children, match_data);
+	      g_list_free (children);
+	    }
+	}
+    }
     
   image = match_theme_image (style, match_data);
   if (image)
@@ -148,7 +280,7 @@ draw_simple_image(GtkStyle       *style,
 	  GdkBitmap *mask = NULL;
 	  gboolean valid;
 
-	  maskwin = get_window_for_shape (image, window, widget);
+	  maskwin = get_window_for_shape (image, window, widget, x, y, width, height);
 	  if (maskwin)
 	    mask = gdk_pixmap_new (maskwin, width, height, 1);
 
@@ -194,17 +326,13 @@ draw_gap_image(GtkStyle       *style,
 	       gint            gap_width)
 {
   ThemeImage *image;
-  gboolean setbg = FALSE;
   
   if ((width == -1) && (height == -1))
-    {
-      gdk_window_get_size(window, &width, &height);
-      setbg = TRUE;
-    }
+    gdk_drawable_get_size (window, &width, &height);
   else if (width == -1)
-    gdk_window_get_size(window, &width, NULL);
+    gdk_drawable_get_size (window, &width, NULL);
   else if (height == -1)
-    gdk_window_get_size(window, NULL, &height);
+    gdk_drawable_get_size (window, NULL, &height);
 
   if (!(match_data->flags & THEME_MATCH_ORIENTATION))
     {
@@ -429,7 +557,7 @@ draw_shadow(GtkStyle     *style,
   match_data.shadow = shadow;
   match_data.state = state;
 
-  if (!draw_simple_image (style, window, area, widget, &match_data, FALSE, FALSE,
+  if (!draw_simple_image (style, window, area, widget, &match_data, FALSE,
 			  x, y, width, height))
     parent_class->draw_shadow (style, window, state, shadow, area, widget, detail,
 			       x, y, width, height);
@@ -529,7 +657,7 @@ draw_arrow (GtkStyle     *style,
       match_data.state = state;
       match_data.arrow_direction = arrow_direction;
       
-      if (draw_simple_image (style, window, area, widget, &match_data, TRUE, TRUE,
+      if (draw_simple_image (style, window, area, widget, &match_data, TRUE,
 			     box_x, box_y, box_width, box_height))
 	{
 	  /* The theme included stepper images, we're done */
@@ -544,12 +672,16 @@ draw_arrow (GtkStyle     *style,
       match_data.shadow = shadow;
       match_data.state = state;
       
-      if (!draw_simple_image (style, window, area, widget, &match_data, TRUE, TRUE,
+      if (!draw_simple_image (style, window, area, widget, &match_data, TRUE,
 			      box_x, box_y, box_width, box_height))
 	parent_class->draw_box (style, window, state, shadow, area, widget, detail,
 				box_x, box_y, box_width, box_height);
     }
 
+#if GTK_CHECK_VERSION(2,9,2)
+  if (arrow_direction == GTK_ARROW_NONE)
+    return;
+#endif
 
   match_data.function = TOKEN_D_ARROW;
   match_data.detail = (gchar *)detail;
@@ -560,7 +692,7 @@ draw_arrow (GtkStyle     *style,
   match_data.state = state;
   match_data.arrow_direction = arrow_direction;
   
-  if (!draw_simple_image (style, window, area, widget, &match_data, TRUE, TRUE,
+  if (!draw_simple_image (style, window, area, widget, &match_data, TRUE,
 			  x, y, width, height))
     parent_class->draw_arrow (style, window, state, shadow, area, widget, detail,
 			      arrow_direction, fill, x, y, width, height);
@@ -590,7 +722,7 @@ draw_diamond (GtkStyle     *style,
   match_data.shadow = shadow;
   match_data.state = state;
   
-  if (!draw_simple_image (style, window, area, widget, &match_data, TRUE, TRUE,
+  if (!draw_simple_image (style, window, area, widget, &match_data, TRUE,
 			  x, y, width, height))
     parent_class->draw_diamond (style, window, state, shadow, area, widget, detail,
 				x, y, width, height);
@@ -634,6 +766,38 @@ draw_string (GtkStyle * style,
     }
 }
 
+static void
+maybe_check_submenu_state (GtkMenuItem *menu_item, ThemeMatchData *match_data)
+{
+  /* Distinguish between active and passive focus, depending on whether the
+   * focus is in submenu.
+   *
+   * Active focus:
+   *   function = BOX
+   *   state    = PRELIGHT
+   *
+   * Passive focus:
+   *   function = BOX
+   *   state    = SELECTED
+   */
+  if (menu_item->submenu)
+    {
+      GtkWidget *sub_item;
+
+      sub_item = GTK_MENU_SHELL (menu_item->submenu)->active_menu_item;
+      if (sub_item && GTK_WIDGET_STATE (sub_item) != GTK_STATE_NORMAL)
+	match_data->state = GTK_STATE_SELECTED;
+    }
+
+  /* This is a hack to cover the cases where a menu item is being 
+   * used outside the usual a GtkMenu. More specificaly, this covers
+   * menu items inside the hildon-desktop custom menu widget. */
+  if (!GTK_IS_MENU (GTK_WIDGET (menu_item)->parent) &&
+      !gtk_widget_is_focus (GTK_WIDGET (menu_item)) &&
+      GTK_WIDGET_STATE (menu_item) != GTK_STATE_NORMAL)
+    match_data->state = GTK_STATE_SELECTED;
+}
+
 static void
 draw_box (GtkStyle     *style,
 	  GdkWindow    *window,
@@ -665,13 +829,63 @@ draw_box (GtkStyle     *style,
   match_data.shadow = shadow;
   match_data.state = state;
 
-  if (!draw_simple_image (style, window, area, widget, &match_data, TRUE, TRUE,
+  if (GTK_IS_MENU_ITEM (widget))
+    maybe_check_submenu_state (GTK_MENU_ITEM (widget), &match_data);
+
+  if (!draw_simple_image (style, window, area, widget, &match_data, TRUE,
 			  x, y, width, height)) {
     parent_class->draw_box (style, window, state, shadow, area, widget, detail,
 			    x, y, width, height);
   }
 }
 
+static void
+maybe_check_cursor_position (GtkTreeView *treeview,
+                             gint x, gint y, gint width, gint height,
+                             ThemeMatchData *match_data)
+{
+  GtkTreePath *cursor_path;
+  GdkRectangle cursor_rect;
+  GdkRectangle paint_rect;
+
+  gtk_tree_view_get_cursor (treeview, &cursor_path, NULL);
+  if (!cursor_path)
+    return;
+
+  /* we only really care about the vertical position here */
+
+  gtk_tree_view_get_background_area (treeview, cursor_path, NULL, &cursor_rect);
+  gtk_tree_path_free (cursor_path);
+
+  paint_rect.y = y;
+  paint_rect.height = height;
+
+  paint_rect.x = cursor_rect.x = x;
+  paint_rect.width = cursor_rect.width = width;
+
+  if (!gdk_rectangle_intersect (&paint_rect, &cursor_rect, &paint_rect))
+    return;
+
+  /* We're painting the cursor row background, so distinguish between focused
+   * and non-focused treeview so that we can paint the passive focus ring when
+   * non-focused. (gtk_paint_focus is used to paint the focus ring when focused.)
+   * Knowing that treeview always uses GTK_SHADOW_NONE, it should be (more or
+   * less) safe to (ab)use it for passive focus.
+   *
+   * Passive focus:
+   *   function = FLAT_BOX
+   *   state    = SELECTED / NORMAL
+   *   shadow   = OUT
+   *
+   * Selection / normal row:
+   *   function = FLAT_BOX
+   *   state    = SELECTED / NORMAL
+   *   shadow   = NONE
+   */
+  if (!GTK_WIDGET_HAS_FOCUS (treeview))
+    match_data->shadow = GTK_SHADOW_OUT;
+}
+
 static void
 draw_flat_box (GtkStyle     *style,
 	       GdkWindow    *window,
@@ -696,7 +910,11 @@ draw_flat_box (GtkStyle     *style,
   match_data.shadow = shadow;
   match_data.state = state;
 
-  if (!draw_simple_image (style, window, area, widget, &match_data, TRUE, TRUE,
+  /* Special handling for treeview cursor row */
+  if (GTK_IS_TREE_VIEW (widget))
+    maybe_check_cursor_position (GTK_TREE_VIEW (widget), x, y, width, height, &match_data);
+
+  if (!draw_simple_image (style, window, area, widget, &match_data, TRUE,
 			  x, y, width, height))
     parent_class->draw_flat_box (style, window, state, shadow, area, widget, detail,
 				 x, y, width, height);
@@ -725,8 +943,16 @@ draw_check (GtkStyle     *style,
   match_data.flags = THEME_MATCH_SHADOW | THEME_MATCH_STATE;
   match_data.shadow = shadow;
   match_data.state = state;
-  
-  if (!draw_simple_image (style, window, area, widget, &match_data, TRUE, TRUE,
+
+  /* Special casing for GtkCheckButton: We want to set the widget state to
+   * ACTIVE to get the correct graphics used in the RC files. Ideally we'd
+   * use the FOCUS rules, but this is not possible due to technical limitations
+   * in how focus is drawn in sapwood */
+  if (GTK_IS_CHECK_BUTTON (widget) &&
+      GTK_WIDGET_HAS_FOCUS (widget))
+    match_data.state = GTK_STATE_ACTIVE;
+
+  if (!draw_simple_image (style, window, area, widget, &match_data, TRUE,
 			  x, y, width, height))
     parent_class->draw_check (style, window, state, shadow, area, widget, detail,
 			      x, y, width, height);
@@ -756,7 +982,15 @@ draw_option (GtkStyle      *style,
   match_data.shadow = shadow;
   match_data.state = state;
   
-  if (!draw_simple_image (style, window, area, widget, &match_data, TRUE, TRUE,
+  /* Special casing for GtkRadioButton: We want to set the widget state to
+   * ACTIVE to get the correct graphics used in the RC files. Ideally we'd
+   * use the FOCUS rules, but this is not possible due to technical limitations
+   * in how focus is drawn in sapwood */
+  if (GTK_IS_RADIO_BUTTON (widget) &&
+      GTK_WIDGET_HAS_FOCUS (widget))
+    match_data.state = GTK_STATE_ACTIVE;
+
+  if (!draw_simple_image (style, window, area, widget, &match_data, TRUE,
 			  x, y, width, height))
     parent_class->draw_option (style, window, state, shadow, area, widget, detail,
 			       x, y, width, height);
@@ -785,8 +1019,8 @@ draw_tab (GtkStyle     *style,
   match_data.flags = THEME_MATCH_SHADOW | THEME_MATCH_STATE;
   match_data.shadow = shadow;
   match_data.state = state;
-  
-  if (!draw_simple_image (style, window, area, widget, &match_data, TRUE, TRUE,
+
+  if (!draw_simple_image (style, window, area, widget, &match_data, TRUE,
 			  x, y, width, height))
     parent_class->draw_tab (style, window, state, shadow, area, widget, detail,
 			    x, y, width, height);
@@ -858,6 +1092,56 @@ draw_box_gap (GtkStyle       *style,
 				x, y, width, height, gap_side, gap_x, gap_width);
 }
 
+static void
+draw_expander (GtkStyle        *style,
+               GdkWindow       *window,
+               GtkStateType     state,
+               GdkRectangle    *area,
+               GtkWidget       *widget,
+               const gchar     *detail,
+               gint             center_x,
+               gint             center_y,
+               GtkExpanderStyle expander_style)
+{
+  ThemeMatchData match_data;
+  gint expander_size = 10;
+
+  g_return_if_fail(style != NULL);
+  g_return_if_fail(window != NULL);
+
+  /* Reusing the arrow theming here as it's flexible enough (almost, we do lose
+   * the intermediate states.) It also allows us to use existing gtkrc.
+   * XXX Might want to introduce proper keywords for expanders some day.
+   */
+
+  if (widget)
+    gtk_widget_style_get (widget, "expander-size", &expander_size, NULL);
+
+  match_data.function = TOKEN_D_ARROW;
+  match_data.detail = (gchar *)detail;
+  match_data.flags = THEME_MATCH_STATE | THEME_MATCH_ARROW_DIRECTION;
+  match_data.state = state;
+
+  switch (expander_style)
+    {
+    case GTK_EXPANDER_COLLAPSED:
+    case GTK_EXPANDER_SEMI_COLLAPSED:
+      match_data.arrow_direction = GTK_ARROW_RIGHT;
+      break;
+    case GTK_EXPANDER_EXPANDED:
+    case GTK_EXPANDER_SEMI_EXPANDED:
+      match_data.arrow_direction = GTK_ARROW_DOWN;
+      break;
+    default:
+      g_return_if_reached ();
+    }
+
+  if (!draw_simple_image (style, window, area, widget, &match_data, TRUE,
+                          center_x - expander_size/2, center_y - expander_size/2, expander_size, expander_size))
+    parent_class->draw_expander (style, window, state, area, widget, detail,
+                                 center_x, center_y, expander_style);
+}
+
 static void
 draw_extension (GtkStyle       *style,
 		GdkWindow      *window,
@@ -884,7 +1168,7 @@ draw_extension (GtkStyle       *style,
   match_data.state = state;
   match_data.gap_side = gap_side;
 
-  if (!draw_simple_image (style, window, area, widget, &match_data, TRUE, TRUE,
+  if (!draw_simple_image (style, window, area, widget, &match_data, TRUE,
 			  x, y, width, height))
     parent_class->draw_extension (style, window, state, shadow, area, widget, detail,
 				  x, y, width, height, gap_side);
@@ -912,7 +1196,7 @@ draw_focus (GtkStyle     *style,
   match_data.flags = THEME_MATCH_STATE;
   match_data.state = state;
   
-  if (!draw_simple_image (style, window, area, widget, &match_data, TRUE, FALSE,
+  if (!draw_simple_image (style, window, area, widget, &match_data, TRUE,
 			  x, y, width, height))
     parent_class->draw_focus (style, window, state, area, widget, detail,
 			      x, y, width, height);
@@ -946,7 +1230,7 @@ draw_slider (GtkStyle      *style,
   match_data.state = state;
   match_data.orientation = orientation;
 
-  if (!draw_simple_image (style, window, area, widget, &match_data, TRUE, TRUE,
+  if (!draw_simple_image (style, window, area, widget, &match_data, TRUE,
 			  x, y, width, height))
     parent_class->draw_slider (style, window, state, shadow, area, widget, detail,
 			       x, y, width, height, orientation);
@@ -981,7 +1265,7 @@ draw_handle (GtkStyle      *style,
   match_data.state = state;
   match_data.orientation = orientation;
 
-  if (!draw_simple_image (style, window, area, widget, &match_data, TRUE, TRUE,
+  if (!draw_simple_image (style, window, area, widget, &match_data, TRUE,
 			  x, y, width, height))
     parent_class->draw_handle (style, window, state, shadow, area, widget, detail,
 			       x, y, width, height, orientation);
@@ -1055,6 +1339,32 @@ render_icon (GtkStyle               *style,
     return parent_class->render_icon (style, source, direction, state, size, widget, detail);
 }
 
+static void
+draw_layout (GtkStyle     *style,
+	     GdkWindow    *window,
+	     GtkStateType  state_type,
+	     gboolean      use_text,
+	     GdkRectangle *area,
+	     GtkWidget    *widget,
+	     const char   *detail,
+	     gint          x,
+	     gint          y,
+	     PangoLayout  *layout)
+{
+  /* Simply draw the text, without any of the fancy effects for insensitive
+   * state in the default theme engine */
+  GdkGC *gc;
+
+  gc = use_text ? style->text_gc[state_type] : style->fg_gc[state_type];
+
+  if (area)
+    gdk_gc_set_clip_rectangle (gc, area);
+
+  gdk_draw_layout (window, gc, x, y, layout);
+
+  if (area)
+    gdk_gc_set_clip_rectangle (gc, NULL);
+}
 
 GType sapwood_type_style = 0;
 
@@ -1110,4 +1420,6 @@ sapwood_style_class_init (SapwoodStyleClass *klass)
   style_class->draw_slider = draw_slider;
   style_class->draw_handle = draw_handle;
   style_class->render_icon = render_icon;
+  style_class->draw_expander = draw_expander;
+  style_class->draw_layout = draw_layout;
 }
