@@ -58,24 +58,24 @@ enum
   TOKEN_D_HLINE,
   TOKEN_D_VLINE,
   TOKEN_D_SHADOW,
-  TOKEN_D_POLYGON,
+  TOKEN_UNUSED_1,
   TOKEN_D_ARROW,
   TOKEN_D_DIAMOND,
-  TOKEN_D_OVAL,
-  TOKEN_D_STRING,
+  TOKEN_UNUSED_2,
+  TOKEN_UNUSED_3,
   TOKEN_D_BOX,
   TOKEN_D_FLAT_BOX,
   TOKEN_D_CHECK,
   TOKEN_D_OPTION,
-  TOKEN_D_CROSS,
-  TOKEN_D_RAMP,
+  TOKEN_UNUSED_4,
+  TOKEN_UNUSED_5,
   TOKEN_D_TAB,
   TOKEN_D_SHADOW_GAP,
   TOKEN_D_BOX_GAP,
   TOKEN_D_EXTENSION,
   TOKEN_D_FOCUS,
   TOKEN_D_SLIDER,
-  TOKEN_D_ENTRY,
+  TOKEN_UNUSED_6,
   TOKEN_D_HANDLE,
   TOKEN_D_STEPPER,
   TOKEN_TRUE,
@@ -98,7 +98,8 @@ enum
   TOKEN_ETCHED_OUT,
   TOKEN_ORIENTATION,
   TOKEN_HORIZONTAL,
-  TOKEN_VERTICAL
+  TOKEN_VERTICAL,
+  TOKEN_POSITION
 };
 
 typedef enum
@@ -120,9 +121,17 @@ typedef enum {
   THEME_MATCH_ORIENTATION     = 1 << 1,
   THEME_MATCH_STATE           = 1 << 2,
   THEME_MATCH_SHADOW          = 1 << 3,
-  THEME_MATCH_ARROW_DIRECTION = 1 << 4
+  THEME_MATCH_ARROW_DIRECTION = 1 << 4,
+  THEME_MATCH_POSITION        = 1 << 5
 } ThemeMatchFlags;
 
+typedef enum {
+  THEME_POS_LEFT   = 1 << 0, /* GTK_POS_LEFT   */
+  THEME_POS_RIGHT  = 1 << 1, /* GTK_POS_RIGHT  */
+  THEME_POS_TOP    = 1 << 2, /* GTK_POS_TOP    */
+  THEME_POS_BOTTOM = 1 << 3  /* GTK_POS_BOTTOM */
+} ThemePositionFlags;
+
 struct _ThemePixbuf
 {
   const char *dirname;
@@ -130,12 +139,12 @@ struct _ThemePixbuf
 
   SapwoodPixmap *pixmap;
 
+  guint16     border_left;
+  guint16     border_right;
+  guint16     border_bottom;
+  guint16     border_top;
   guint       refcnt : 7;
   guint       stretch : 1;
-  guint8      border_left;
-  guint8      border_right;
-  guint8      border_bottom;
-  guint8      border_top;
 };
 
 struct _ThemeMatchData
@@ -143,11 +152,12 @@ struct _ThemeMatchData
   gchar          *detail;
   guint16         function;	/* Mandatory */
 
-  ThemeMatchFlags flags           : 5;
+  ThemeMatchFlags flags           : 6;
+  ThemePositionFlags position     : 4;
   GtkStateType    state           : 3;
   GtkShadowType   shadow          : 3;
   GtkPositionType gap_side        : 2;
-  GtkArrowType    arrow_direction : 2;
+  guint           arrow_direction : 2;	/* GtkArrowType, but without NONE */
   GtkOrientation  orientation     : 1;
 };
 
@@ -166,20 +176,20 @@ struct _ThemeImage
 };
 
 
-ThemePixbuf *theme_pixbuf_new          (void);
-void         theme_pixbuf_unref        (ThemePixbuf  *theme_pb);
+ThemePixbuf *theme_pixbuf_new          (void) G_GNUC_INTERNAL;
+void         theme_pixbuf_unref        (ThemePixbuf  *theme_pb) G_GNUC_INTERNAL;
 void         theme_pixbuf_set_filename (ThemePixbuf  *theme_pb,
-					const char   *filename);
+					const char   *filename) G_GNUC_INTERNAL;
 gboolean     theme_pixbuf_get_geometry (ThemePixbuf  *theme_pb,
 					gint         *width,
-					gint         *height);
+					gint         *height) G_GNUC_INTERNAL;
 void         theme_pixbuf_set_border   (ThemePixbuf  *theme_pb,
 					gint          left,
 					gint          right,
 					gint          top,
-					gint          bottom);
+					gint          bottom) G_GNUC_INTERNAL;
 void         theme_pixbuf_set_stretch  (ThemePixbuf  *theme_pb,
-					gboolean      stretch);
+					gboolean      stretch) G_GNUC_INTERNAL;
 gboolean     theme_pixbuf_render       (ThemePixbuf  *theme_pb,
 					GdkWindow    *window,
 					GdkBitmap    *mask,
@@ -189,8 +199,8 @@ gboolean     theme_pixbuf_render       (ThemePixbuf  *theme_pb,
 					gint          dest_x,
 					gint          dest_y,
 					gint          dest_width,
-					gint          dest_height);
+					gint          dest_height) G_GNUC_INTERNAL;
 
 
 
-extern GtkStyleClass pixmap_default_class;
+extern GtkStyleClass pixmap_default_class G_GNUC_INTERNAL;
